<?php

/* Copyright (C) 2010 Julien VAUBOURG <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Tested with Centreon 2.0.2
 *
 */

# error_reporting(E_ALL);
# ini_set('display_errors', 1);

define('OK', 0);
define('WARNING', 1);
define('CRITICAL', 2);

$states = array('ok', 'warning', 'critical');

// Centreon server
$mysql_server = '';

// The user must have the select right on all tables of nagios and centreon2
$mysql_user = '';
$mysql_pwd = '';

// Every $refresh seconds, the page will refreshed
$refresh = 5 * 60;

// Allow debug mode, activated by add ?debug GET parameter to the URL
$allowDebugMode = false;


/********************************
 *********** FUNCTIONS **********
 ********************************/

/*
 * Read the cache file corresponding to the conf file used
 * and return a globalState array version, completed
 * (see its definition in the main program for consult the structure)
 *
 * Cache file syntax (with line numbers) :
 * 0: timestampOfTheLastCachRefresh
 * 1: titlePage
 * [n: ID|label|description|status|timestampOfTheLastUpdate
 * 	...]
 */ 
function readCache($file) {
	$lastRefresh = false;
	$globalStates = array();

	// There is an existing cache file for the conf file used
	if(file_exists($file)) {

		// Opening cache file
		$fcache = fopen($file, 'r')
			or die("ERROR: The cache file can't be read");

		// The first line is the unix timestamp of the last cache refresh
		$line = fgets($fcache, 1024);
		$lastRefresh = trim($line);

		// The second line is the page title
		$line = fgets($fcache, 1024);
		$page_title = trim($line);

		$l = 2;

		// For each global service of the cache file
		while($line = fgets($fcache, 1024)) {
			$l++;

			// See comments' function for consult the lines structure
			list($id, $label, $description, $state, $time) =
				explode('|', trim($line));

			(empty($id) || empty($label) || empty($state) && $state != 0 || empty($time))
				&& die("ERROR (cache l.$l) : Wrong syntax or missing field");

			// Filling the main array
			$globalStates[$id]['label'] = $label;
			$globalStates[$id]['description'] = $description;
			$globalStates[$id]['state'] = $state;
			$globalStates[$id]['newState'] = OK;
			$globalStates[$id]['time'] = $time;
			$globalStates[$id]['newTime'] = false;
		}

		fclose($fcache);

	// There is no cache file
	} else {
		$page_title = NULL;
		$lastRefresh = -1;
		$globalState = array();
	}

	return array($lastRefresh, $page_title, $globalStates);
}

/*
 * Callback function, for delete pipes in the cache fields
 * (pipes are used as delimiters)
 */
function dropPipes(&$value, $key) {
	$value = str_replace('|', NULL, $value);
}

/*
 * Update the main array globalStates, and the cache file
 */
function updateCacheAndValues($file, $lastRefresh, $page_title, &$globalStates) {

	// New refresh date
	$cache = time()."\n$page_title\n";

	// For each global service
	while($id = key($globalStates)) {
		$infos =& $globalStates[$id];

		// The update time is refreshed only if the state was changed, and if the old date is
		// really the oldest (wich should alway be the case)
		if($infos['state'] != $infos['newState'] && $infos['time'] < $infos['newTime']) {
			$infos['state'] = $infos['newState'];
			$infos['time'] = $infos['newTime'];
		}

		// Delete pipe chars, used as delimiters
		array_walk($infos, 'dropPipes');

		// New cache line, updated
		$cache .= sprintf("%s|%s|%s|%s|%s\n",
				$id,
				$infos['label'],
				$infos['description'],
				$infos['state'],
				$infos['time']
			);

		next($globalStates);
	}

	file_put_contents($file, $cache)
		|| die("ERROR: The cache file can't be written");
}

/*
 * Extends hostgroups, prefixed by HG: :
 * 	If the line contains HG:my-hg!my-srvce, and my-hg has 3 hosts, it'll be replaced by :
 * 	my-host1!my-srvce my-host2!my-srvce my-host3!my-srvce
 */
function extendHG($line, $l) {

	// For each hostgroup definition found
	while(preg_match('/HG:([^!\s]+)!?(\S+)?/', $line, $hg)) {
		$hosts = array();

		// This hosgroup definition has an associated service ?
		(count($hg) > 2) ?
			list($hg_def, $hg_name, $hg_service) = $hg :
			list($hg_def, $hg_name) = $hg;

		// Select all hosts that it belong to the hostgroup
		$req = mysql_query(
			sprintf("SELECT host_name
					FROM centreon2.hostgroup_relation, centreon2.host
					WHERE host_host_id = host_id
					AND hostgroup_hg_id=(SELECT hg_id
								FROM centreon2.hostgroup
								WHERE hg_name='%s'
								LIMIT 1)",
				mysql_real_escape_string($hg_name)
			));

		// This hosgroup really exist ?
		mysql_num_rows($req)
			|| die("ERROR (conf l.$l) : Hostgroup <em>$hg_name</em> does not exist or is empty");

		// Each host, eventually associated to a service, is concatenated
		while($host = mysql_fetch_assoc($req))
			$hosts[] = $host['host_name'].(isset($hg_service) ? "!$hg_service" : NULL);

		// The hostgroup definition is replaced by its members
		$line = str_replace($hg_def, implode(' ', $hosts), $line);
	}

	// Line definition, without hostgroup definitions
	return $line;
}

/*
 * Replace the percentages by absolute numbers
 */
function replacePercent($percent, $nbServices) {
	if(preg_match('/(\d+)%$/', $percent, $nb))
		$percent = $nbServices / 100 * $nb[1];

	return $percent;
}

/*
 * Return a relative date
 * Adapted from :
 *	http://blog.angechierchia.com/php-mysql/afficher-une-date-relative-en-php
 */
function getRelativeTime($time) {
	$timeDiff = time() - $time; 

	if($timeDiff <= 0)
		return "moins d'une seconde";

	$timeDiff = abs($timeDiff);

	$times = array(
		31104000 => 'an{s}',       // 12 * 30 * 24 * 60 * 60 secondes
		2592000  => 'mois',        // 30 * 24 * 60 * 60 secondes
		604800   => 'semaine{s}',  // 7 * 24 * 60 * 60 secondes
		86400    => 'jour{s}',     // 24 * 60 * 60 secondes
		3600     => 'heure{s}',    // 60 * 60 secondes
		60       => 'minute{s}');

	$strTime = NULL;

	// Until that the rest can't being converted
	while($timeDiff >= 60) {
		foreach($times AS $seconds => $unit) {
			$delta = floor($timeDiff / $seconds);

			if($delta >= 1) {
				$unit = str_replace('{s}', ($delta == 1 ? NULL : 's'), $unit);
				$strTime .= "$delta $unit ";
				$timeDiff -= $delta * $seconds;
			}
		}
	}

	// If there is still seconds
	($timeDiff > 0)
		&& $strTime .= "et $timeDiff secondes";

	return trim($strTime); 
}



/********************************
 ************* MAIN *************
 ********************************/

// ID of the global service that is parsed
$currentDefID = false;

// Debug mode activated (this requires to have $allowDebugMode on true)
if(isset($_GET['debug'])) {
	$debug = true;
	unset($_GET['debug']);
} else
	$debug = false;

// With ?id=<ID> in the URL, you can have only the traffic lights for only one service
if(isset($_GET['id'])) {
	$serviceID = strtolower($_GET['id']);
	unset($_GET['id']);
} else
	$serviceID = false;

// With ?id=<ID>&rotate, you can have the traffic lights horizontally
if(isset($_GET['rotate'])) {
	$rotate = true;
	unset($_GET['rotate']);
} else
	$rotate = false;

// If this page is called with ?toto, the conf file confs/conf_toto will be used
// By default, it'll be confs/conf_public
$confID = empty($_GET) ? 'public' : key($_GET);

// Opening conf file
$fconf = fopen("./confs/conf_$confID", 'r')
	or die("ERROR: The conf file can't be read");

// Global states hash structure :
// 'ID' => [
//	'label' => AAAA
//	'state' => X
//	'description' => AAAA
//	'time' => XXXX
//	'newTime' => XXXX
// ]
list(
	$lastRefresh, // Last time the cache has been updated
	$page_title, // Main page title
	$globalStates // See above
) = readCache("./caches/cache_$confID");

$l = 0;

// Cache's infos are outdated
if(time() - $lastRefresh > $refresh) {

	// Connection to the Centreon DB
	mysql_connect($mysql_server, $mysql_user, $mysql_pwd)
		|| die ("ERROR: Mysql connect");

	// For each line of the conf file
	while($line = fgets($fconf, 1024)) {
		$l++;

		// A line beginning by PAGE_TITLE is the line that define the page title
		if(preg_match('/^\s*PAGE_TITLE\s*=\s*(.*)/', $line, $title)) {
			$page_title = htmlentities(utf8_decode($title[1]));

		// A white line is a separator between both global service definition
		} elseif(preg_match('/^\s*$/', $line))
			$currentDefID = false;

		// All lines except comments
		elseif(!preg_match('/^\s*#/', $line)) {

			// Drop white spaces and CR
			$line = trim($line);

			// This line is not part of the current definition
			// (it's the begin of a new definition)
			if(!$currentDefID) {
				$description = NULL;
				$refID = false;

				// If there is an ID found for this global service, it's recovered
				// and deleted
				if(preg_match('/\s*ID\s*=\s*([a-z0-9_-]+)\s*$/i', $line, $id)) {
					$refID = strtolower($id[1]);
					$line = str_replace($id[0], NULL, $line);
				} else
					die("ERROR (conf l.$l) : No ID found");

				// This global service was not in the cache
				!isset($globalStates[$refID])
					&& $globalStates[$refID] = array(
						'description' => NULL,
						'state' => -1,
						'newState' => OK,
						'time' => -1,
						'newTime' => -1
					);

				// If there is a ! delimiter in the label, this means that there is
				// a description attached (label ! description)
				if(strpos($line, '!')) {
					list($label, $description) = preg_split('/\s*!\s*/', $line);

					$globalStates[$refID]['label'] = $label;
					$globalStates[$refID]['description'] = $description;
				} else
					$globalStates[$refID]['label'] = $line;

				$currentDefID = $refID;

			// This line is part of the current definition.
			// It's considered only if the global service state is not already critical.
			} elseif($globalStates[$currentDefID]['newState'] != CRITICAL) {

				// Change hostgroup definitions by its hosts
				$line = extendHG($line, $l);

				// Drop the optional linking words
				$line = preg_replace('/\s*(OR|AND)\s+/', ' ', $line);

				// Split each host[!service] definition
				$orConditions = preg_split('/[\s]+/', trim($line));

				// By default, for have a critical state for the line,
				// it should that at least one host!service is critical
				$nbCrit4Crit = $nbCrit4Warn = count($orConditions);

				// Last condition is either a bounds parameter, or a classical condition
				$lastCondition = array_pop($orConditions);

				// Last condition is a bounds parameter (!nbCritForWarn!nbCritForCrit)
				if(preg_match('/^!(\d+%?)!?(\d+%?)?$/', $lastCondition, $limits)) {

					// There is two bounds ? (warning and critical)
					(count($limits) > 2) ?
						list(, $nbCrit4Warn, $nbCrit4Crit) = $limits : // Warn and crit are extracted
						$nbCrit4Crit = $nbCrit4Warn = $limits[1]; // Crit is extracted, and warn is overloaded with its value

					// Replace the bounds percentages by absolute values
					$nbServices = count($orConditions);
					$nbCrit4Warn = replacePercent($nbCrit4Warn, $nbServices);
					$nbCrit4Crit = replacePercent($nbCrit4Crit, $nbServices);

				// Last condition is really a condition, it's re-pushed in the array
				} else
					array_push($orConditions, $lastCondition);

				// The critical bound is necessarily higher than the warn bound
				($nbCrit4Crit >= $nbCrit4Warn)
					|| die("ERROR (conf l.$l) : Warn bound is lower than crit bound (W:$nbCrit4Warn, C:$nbCrit4Crit)");

				// Nb critical services (or down hosts) in the line
				$nbCrit = 0;
				$req = false;

				// For each condition : what's its state ?
				foreach($orConditions AS $orCondition) {

					// If the ! delimiter is found, there is a service associated with a host
					// (else, there is an host alive test)
					$isCheckAlive = !strpos($orCondition, '!');
					$host = $service = false;

					// It's an host alive test
					if($isCheckAlive) {
						$host = $orCondition;

						// Recover the state of the host
						// (0 or 1 => in this case 1 is therefore critical/down, and not warning)
						$req = mysql_query(
							sprintf("SELECT nhs.last_hard_state, UNIX_TIMESTAMP(nhs.last_hard_state_change)
									FROM nagios.nagios_hoststatus nhs, nagios.nagios_objects no
									WHERE nhs.host_object_id = no.object_id
									AND no.name1 = '%s'
									LIMIT 1",
								mysql_real_escape_string($host)
							));

						mysql_num_rows($req)
							|| die("ERROR (conf l.$l) : No state found for <em>$host</em>");

					// It's a service associated with a host
					} else {

						// Separate host and service
						list($host, $service) = explode('!', $orCondition);

						// Recover the of the service for this host (0, 1 or 2)
						$req = mysql_query(
							sprintf("SELECT nss.last_hard_state, UNIX_TIMESTAMP(nss.last_hard_state_change)
									FROM nagios.nagios_servicestatus nss, nagios.nagios_objects no, nagios.nagios_services ns
									WHERE no.object_id = nss.service_object_id
									AND no.object_id = ns.service_object_id
									AND no.name1='%s'
									AND no.name2='%s'
									LIMIT 1",
								mysql_real_escape_string($host),
								mysql_real_escape_string($service)
							));

						mysql_num_rows($req)
							|| die("ERROR (conf l.$l) : No relation found between <em>$host</em> and <em>$service</em>");
					}

					list($state, $lastChange) = mysql_fetch_row($req);

					// For a host alive test, anything that is higher than ok (0) is critical (2)
					if($isCheckAlive)
						$state > OK && $state = CRITICAL;

					// Else, anything that is not OK, warning or critical is ok (i.e. unknown state)
					else
						($state < OK || $state == WARNING || $state > CRITICAL)
							&& $state = OK;

					// Display state for each service, in the debug mode
					$allowDebugMode && isset($_GET['debug'])
						&& printf("DEBUG : <strong>%s</strong>!%s => <em>%s</em><br />\n",
							$host,
							$service ? $service : 'alive',
							strtoupper($states[$state]));

					// If this service/host is critical, it's a new critical for this line
					($state == CRITICAL)
						&& $nbCrit++;

					// Save the newest service state, for this global service
					($globalStates[$currentDefID]['newTime'] < $lastChange)
						&& $globalStates[$currentDefID]['newTime'] = $lastChange;
				}

				// If the nb of critical is higher than the critical bound, and the global service state is not already
				// critical, this one becomes critical
				if($nbCrit >= $nbCrit4Crit && $globalStates[$currentDefID]['newState'] < CRITICAL)
					$globalStates[$currentDefID]['newState'] = CRITICAL;

				// If the nb of critical is higher than the warning bound, and the global service state is not higher
				// than warning (therefore ok...), this one becomes warning
				elseif($nbCrit >= $nbCrit4Warn && $globalStates[$currentDefID]['newState'] < WARNING)
					$globalStates[$currentDefID]['newState'] = WARNING;
			}
		}
	}

	// The End
	fclose($fconf);
	mysql_close();

	// Update the cache file and $globalStates values
	updateCacheAndValues("./caches/cache_$confID", $lastRefresh, $page_title, $globalStates);

// All values are recovered from the cache file
} elseif($allowDebugMode && isset($_GET['debug']))
	echo "DEBUG : The cache file is used";


$favicon = OK;

// If a least one global service is critical, the favicon (traffic lights) displays critical.
// Else, if a least one global service is warning, the favicon is warning, else it's ok.
foreach($globalStates AS $infos) {
	if($infos['state'] == CRITICAL) {
		$favicon = CRITICAL;
		break;

	} elseif($infos['state'] == WARNING)
		$favicon = WARNING;
}


/********************************
 ************* HTML *************
 ********************************/

// The user want the full page
if(!$serviceID) : ?>

	<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
	<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="fr" lang="fr">
	<head>
	<title><?php echo $page_title ?></title>
	<meta name="author" content="Julien VAUBOURG / CRI-SRT" />
	<meta name="generator" content="vim -p index.{php,css}"/>
	<meta http-equiv="content-type" content="text/html; charset=utf-8" />
	<meta http-equiv="refresh" content="<?php echo $refresh ?>" />
	<link href="img/<?php echo $states[$favicon] ?>.ico" rel="icon" type="image/x-icon" />
	<link rel="stylesheet" type="text/css" href="index.css" title="Default" />
	</head>

	<body>
	<h1><?php echo $page_title ?></h1>
	<h2>Etat des services</h2>

	<div id="body">
	<?php
		$explanations = Array(
			"Le service est parfaitement fonctionnel.",

			"Le service est perturbé.
			Nous mettons tout en oeuvre pour résoudre ces désagréments.",

			"Le service n'est pas fonctionnel.
			Nous mettons tout en oeuvre pour le rétablir au plus vite."
		);

		foreach($globalStates AS $id => $infos) {
			$relativeTime = getRelativeTime($infos['time']);

			echo <<<HTML
				<div class="state state_{$states[$infos['state']]}">
					<div class="label"><a href="http://contrats.cri.inpl-nancy.fr/$id.html" title="Voir le contrat de service">{$infos['label']}</a></div>
					<div class="description"><span>{$infos['description']}</span></div>

					<div class="explanation">
						{$explanations[$infos['state']]}
					</div>

					<div class="since">Depuis $relativeTime.</div>
				</div>

HTML;
		}
	?>
	</div>

	<div id="copy">
		<a href="" title="GLPI">Ouvrir un ticket</a> -
		<a href="mailto:" title="Service proximité">Contacter le service informatique</a><br />

		<em>
			Dernier rafraîchissement le <?php echo date('d/m/y à H:i:s') ?> 
			(toutes les <?php echo ($refresh % 60) ? "$refresh secondes" : ($refresh / 60).' minutes' ?>)
		</em>
	</div>
	</body>
	</html>


<?php // The user want only one global service id state (traffic lights)
else :

	// No global service with this ID found
	isset($globalStates[$serviceID])
		|| die("ERROR : Global service ID not found");

	$img = sprintf('img/%s%s.png',
		$states[$globalStates[$serviceID]['state']],
		$rotate ? '_h' : NULL);

	header('Content-Length: '.filesize($img));
	header('Content-Type: image/png');

	// Display image with type/mime PNG
	exit(file_get_contents($img));

endif ?>
